module image_processing(
    input VGA_CLK,
    input [4:0] SW,
    input [29:0] datain_x0y0,
    input [29:0] datain_x1y0,
    input [29:0] datain_x2y0,
    input [29:0] datain_x0y1,
    input [29:0] datain_x1y1,
    input [29:0] datain_x2y1,
    input [29:0] datain_x0y2,
    input [29:0] datain_x1y2,
    input [29:0] datain_x2y2,
    output reg [29:0] RGB_out
);

// Constants for color component weights
localparam WEIGHT_R = 100;
localparam WEIGHT_G = 95;
localparam WEIGHT_B = 82;

// Constants for luminosity calculation
localparam LUMA_R_WEIGHT = 30;
localparam LUMA_G_WEIGHT = 60;
localparam LUMA_B_WEIGHT = 11;

// Other constants
localparam GRAY_SHIFT = 2;
localparam SKIN_U_MIN = 40;
localparam SKIN_U_MAX = 296;

reg [9:0] rrsepia;
reg [9:0] ggsepia;
reg [9:0] bbsepia;
reg [19:0] lumir;
reg [19:0] lumig;
reg [19:0] lumib;
reg [16:0] lumin;
reg [39:0] Gx;
reg [39:0] Gy;
reg [39:0] Gn;
reg [9:0] prom;
reg [9:0] gris;
reg [9:0] Y;
reg [9:0] U;
reg [9:0] V;
reg [29:0] skin;


// New Image Processing Filters
reg [19:0] box_blur_r;
reg [19:0] box_blur_g;
reg [19:0] box_blur_b;
reg [9:0] gaussian_blur_r;
reg [9:0] gaussian_blur_g;
reg [9:0] gaussian_blur_b;
reg [9:0] emboss_r;
reg [9:0] emboss_g;
reg [9:0] emboss_b;
reg [9:0] sharpen_r;
reg [9:0] sharpen_g;
reg [9:0] sharpen_b;
reg [9:0] high_pass_r;
reg [9:0] high_pass_g;
reg [9:0] high_pass_b;


// Macro for RGB component selection
`define RGB_COMPONENT(component) {component, component, component}

// Macro for 3x3 kernel calculation
`define KERNEL_3X3(p00, p01, p02, p10, p11, p12, p20, p21, p22) (p00 + p01 + p02 + p10 + p11 + p12 + p20 + p21 + p22) / 9

always @(*)
begin
    // Calculate sepia components
    rrsepia = (datain_x0y0[29:20] * WEIGHT_R) / 100;
    ggsepia = (datain_x0y0[19:10] * WEIGHT_G) / 100;
    bbsepia = (datain_x0y0[9:0] * WEIGHT_B) / 100;

    // Calculate luminosity
    lumir = datain_x0y0[29:20] * LUMA_R_WEIGHT;
    lumig = datain_x0y0[19:10] * LUMA_G_WEIGHT;
    lumib = datain_x0y0[9:0] * LUMA_B_WEIGHT;
    lumin = lumir + lumig + lumib;

    // Calculate Gx and Gy for edge detection
    Gx = (datain_x2y0 + 2 * datain_x2y1 + datain_x2y2) - (datain_x0y0 + 2 * datain_x0y1 + datain_x0y2);
    Gy = (datain_x0y0 + 2 * datain_x1y0 + datain_x2y0) - (datain_x0y2 + 2 * datain_x1y2 + datain_x2y2);

    // Calculate Gn (absolute value of Gx and Gy)
    if (Gx[9] == 0 && Gy[9] == 0) begin
        Gn = Gx + Gy;
	end
    else if (Gx[9] == 1 && Gy[9] == 0)
	begin
        Gn = (~Gx) + Gy;
	end
    else if (Gx[9] == 0 && Gy[9] == 1)
	begin
        Gn = Gx + (~Gy);
	end
    else begin
        Gn = (~Gx) + (~Gy);
	end
    
    // Calculate YUV components
    Y = (datain_x0y0[29:20] + (2 * datain_x0y0[19:10]) + datain_x0y0[9:0]) / 4;
    U = datain_x0y0[29:20] - datain_x0y0[19:10];
    V = datain_x0y0[9:0] - datain_x0y0[19:10];

    // Skin detection based on YUV components
    if (U >= SKIN_U_MIN && U <= SKIN_U_MAX)
	begin
        skin = datain_x0y0;
	end
    else begin
        skin = 30'd0;
	end



	// Calculate Box Blur Filter
    box_blur_r = (datain_x0y0[29:10] + datain_x1y0[29:10] + datain_x2y0[29:10] + datain_x0y1[29:10] + datain_x1y1[29:10] + datain_x2y1[29:10] + datain_x0y2[29:10] + datain_x1y2[29:10] + datain_x2y2[29:10]) >> 3;
    box_blur_g = (datain_x0y0[19:0] + datain_x1y0[19:0] + datain_x2y0[19:0] + datain_x0y1[19:0] + datain_x1y1[19:0] + datain_x2y1[19:0] + datain_x0y2[19:0] + datain_x1y2[19:0] + datain_x2y2[19:0]) >> 3;
    box_blur_b = (datain_x0y0[9:0] + datain_x1y0[9:0] + datain_x2y0[9:0] + datain_x0y1[9:0] + datain_x1y1[9:0] + datain_x2y1[9:0] + datain_x0y2[9:0] + datain_x1y2[9:0] + datain_x2y2[9:0]) >> 3;

    // Calculate Gaussian Blur Filter (using 3x3 kernel)
    gaussian_blur_r = (datain_x0y0[29:0] + 2 * (datain_x1y0[29:0] + datain_x0y1[29:0] + datain_x2y0[29:0] + datain_x0y2[29:0]) + 4 * (datain_x1y1[29:0]) + datain_x2y1[29:0] + datain_x1y2[29:0] + datain_x2y2[29:0]) >> 6;
    gaussian_blur_g = (datain_x0y0[29:0] + 2 * (datain_x1y0[29:0] + datain_x0y1[29:0] + datain_x2y0[29:0] + datain_x0y2[29:0]) + 4 * (datain_x1y1[29:0]) + datain_x2y1[29:0] + datain_x1y2[29:0] + datain_x2y2[29:0]) >> 6;
    gaussian_blur_b = (datain_x0y0[29:0] + 2 * (datain_x1y0[29:0] + datain_x0y1[29:0] + datain_x2y0[29:0] + datain_x0y2[29:0]) + 4 * (datain_x1y1[29:0]) + datain_x2y1[29:0] + datain_x1y2[29:0] + datain_x2y2[29:0]) >> 6;

    // Calculate Emboss Filter
    emboss_r = (datain_x0y0[29:0] - datain_x0y2[29:0] + datain_x2y0[29:0] - datain_x2y2[29:0]) >> 1;
    emboss_g = (datain_x0y0[29:0] - datain_x0y2[29:0] + datain_x2y0[29:0] - datain_x2y2[29:0]) >> 1;
    emboss_b = (datain_x0y0[29:0] - datain_x0y2[29:0] + datain_x2y0[29:0] - datain_x2y2[29:0]) >> 1;

    // Calculate Sharpening Filter
    sharpen_r = datain_x0y0[29:0] + 2 * (datain_x0y1[29:0] + datain_x1y0[29:0] + datain_x1y2[29:0] + datain_x2y1[29:0]) - 6 * datain_x1y1[29:0];
    sharpen_g = datain_x0y0[29:0] + 2 * (datain_x0y1[29:0] + datain_x1y0[29:0] + datain_x1y2[29:0] + datain_x2y1[29:0]) - 6 * datain_x1y1[29:0];
    sharpen_b = datain_x0y0[29:0] + 2 * (datain_x0y1[29:0] + datain_x1y0[29:0] + datain_x1y2[29:0] + datain_x2y1[29:0]) - 6 * datain_x1y1[29:0];

    // Calculate High-Pass Filter (Edge Enhancement)
    high_pass_r = 5 * datain_x1y1[29:0] - datain_x0y0[29:0] - datain_x1y0[29:0] - datain_x2y0[29:0] - datain_x0y1[29:0] - datain_x0y2[29:0] - datain_x1y2[29:0] - datain_x2y2[29:0];
    high_pass_g = 5 * datain_x1y1[29:0] - datain_x0y0[29:0] - datain_x1y0[29:0] - datain_x2y0[29:0] - datain_x0y1[29:0] - datain_x0y2[29:0] - datain_x1y2[29:0] - datain_x2y2[29:0];
    high_pass_b = 5 * datain_x1y1[29:0] - datain_x0y0[29:0] - datain_x1y0[29:0] - datain_x2y0[29:0] - datain_x0y1[29:0] - datain_x0y2[29:0] - datain_x1y2[29:0] - datain_x2y2[29:0];
    // Filter selection based on SW
    case (SW)
        5'd0: RGB_out = {datain_x0y0[29:20], 20'd0}; // Component R
        5'd1: RGB_out = {10'd0, datain_x0y0[19:10], 10'd0}; // Component G
        5'd2: RGB_out = {20'd0, datain_x0y0[9:0]}; // Component B
        5'd3: RGB_out = {datain_x0y0[29:20], datain_x0y0[29:20], datain_x0y0[29:20]}; // Grayscale based on R
        5'd4: RGB_out = {datain_x0y0[19:10], datain_x0y0[19:10], datain_x0y0[19:10]}; // Grayscale based on G
        5'd5: RGB_out = {datain_x0y0[9:0], datain_x0y0[9:0], datain_x0y0[9:0]}; // Grayscale based on B
        5'd6: RGB_out = {lumin[16:7], lumin[16:7], lumin[16:7]}; // Luminosity
        5'd7: RGB_out = {~datain_x0y0[29:20], ~datain_x0y0[19:10], ~datain_x0y0[9:0]}; // Negative
        5'd8: RGB_out = {rrsepia[9:0], ggsepia[9:0], bbsepia[9:0]}; // Sepia
        5'd9: RGB_out = {Gn[9:0], Gn[9:0], Gn[9:0]}; // Edge Detection (Gn)
        5'd10: RGB_out = (datain_x0y0[29:20] < 550) ? {lumin[16:7], lumin[16:7], lumin[16:7]} : datain_x0y0; // Thresholding
        5'd11: RGB_out = (datain_x0y0[19:10] < 550) ? {lumin[16:7], lumin[16:7], lumin[16:7]} : datain_x0y0; // Thresholding (G component)
        5'd12: RGB_out = (datain_x0y0[9:0] < 550) ? {lumin[16:7], lumin[16:7], lumin[16:7]} : datain_x0y0; // Thresholding (B component)
        5'd13: RGB_out = `RGB_COMPONENT(`KERNEL_3X3(datain_x0y0, datain_x1y0, datain_x2y0, datain_x0y1, datain_x1y1, datain_x2y1, datain_x0y2, datain_x1y2, datain_x2y2)); // Average Filter
        5'd14: RGB_out = skin; // Skin Detection

		  5'd15: RGB_out = {box_blur_r[19:10], box_blur_g[19:0], box_blur_b[9:0]}; // Box Blur Filter
        5'd16: RGB_out = {gaussian_blur_r[9:0], gaussian_blur_g[9:0], gaussian_blur_b[9:0]}; // Gaussian Blur Filter
        5'd17: RGB_out = {emboss_r[9:0], emboss_g[9:0], emboss_b[9:0]}; // Emboss Filter
        5'd18: RGB_out = {sharpen_r[9:0], sharpen_g[9:0], sharpen_b[9:0]}; // Sharpening Filter
        5'd19: RGB_out = {high_pass_r[9:0], high_pass_g[9:0], high_pass_b[9:0]}; // High-Pass Filter
        default: RGB_out = datain_x0y0; // Default case
    endcase
end

endmodule
